@@ -38,6 +38,23 @@ except ImportError:  # Python 2.6
 
 logging.getLogger(__name__).addHandler(NullHandler())
 
+
+class UrllibWarningFilter(object):
+    def filter(self, record):
+        return record.msg != "Connection pool is full, discarding connection: %s"
+
+
+logging.getLogger('urllib3.connectionpool').addFilter(UrllibWarningFilter())
+
 import b2sdk.version
 __version__ = b2sdk.version.VERSION
 assert __version__  # PEP-0396
+
+# https://github.com/crsmithdev/arrow/issues/612 - To get rid of the ArrowParseWarning messages in 0.14.3 onward.
+try:
+    from arrow.factory import ArrowParseWarning
+except ImportError:
+    pass
+else:
+    import warnings
+    warnings.simplefilter("ignore", ArrowParseWarning)
