@@ -11,6 +11,7 @@
 from __future__ import print_function
 
 import base64
+import io
 import os
 import random
 import re
@@ -53,8 +54,9 @@ API_VERSION = 'v2'
 
 @unique
 class MetadataDirectiveMode(Enum):
-    COPY = 401
-    REPLACE = 402
+    """ Mode of handling metadata when copying a file """
+    COPY = 401  #: copy metadata from the source file
+    REPLACE = 402  #: ignore the source file metadata and set it to provided values
 
 
 @six.add_metaclass(ABCMeta)
@@ -63,10 +65,70 @@ class AbstractRawApi(object):
     Direct access to the B2 web apis.
     """
 
+    @abstractmethod
+    def authorize_account(self, realm_url, application_key_id, application_key):
+        pass
+
     @abstractmethod
     def cancel_large_file(self, api_url, account_auth_token, file_id):
         pass
 
+    @abstractmethod
+    def copy_file(
+        self,
+        api_url,
+        account_auth_token,
+        source_file_id,
+        new_file_name,
+        bytes_range=None,
+        metadata_directive=None,
+        content_type=None,
+        file_info=None,
+        destination_bucket_id=None,
+    ):
+        pass
+
+    @abstractmethod
+    def copy_part(
+        self,
+        api_url,
+        account_auth_token,
+        source_file_id,
+        large_file_id,
+        part_number,
+        bytes_range=None,
+    ):
+        pass
+
+    @abstractmethod
+    def create_bucket(
+        self,
+        api_url,
+        account_auth_token,
+        account_id,
+        bucket_name,
+        bucket_type,
+        bucket_info=None,
+        cors_rules=None,
+        lifecycle_rules=None
+    ):
+        pass
+
+    @abstractmethod
+    def create_key(
+        self, api_url, account_auth_token, account_id, capabilities, key_name,
+        valid_duration_seconds, bucket_id, name_prefix
+    ):
+        pass
+
+    @abstractmethod
+    def download_file_from_url(self, account_auth_token_or_none, url, range_=None):
+        pass
+
+    @abstractmethod
+    def delete_key(self, api_url, account_auth_token, application_key_id):
+        pass
+
     @abstractmethod
     def delete_bucket(self, api_url, account_auth_token, account_id, bucket_id):
         pass
@@ -79,6 +141,20 @@ class AbstractRawApi(object):
     def finish_large_file(self, api_url, account_auth_token, file_id, part_sha1_array):
         pass
 
+    @abstractmethod
+    def get_download_authorization(
+        self, api_url, account_auth_token, bucket_id, file_name_prefix, valid_duration_in_seconds
+    ):
+        pass
+
+    @abstractmethod
+    def get_file_info(self, api_url, account_auth_token, file_id):
+        pass
+
+    @abstractmethod
+    def get_upload_url(self, api_url, account_auth_token, bucket_id):
+        pass
+
     @abstractmethod
     def get_upload_part_url(self, api_url, account_auth_token, file_id):
         pass
@@ -87,13 +163,66 @@ class AbstractRawApi(object):
     def hide_file(self, api_url, account_auth_token, bucket_id, file_name):
         pass
 
+    @abstractmethod
+    def list_buckets(
+        self,
+        api_url,
+        account_auth_token,
+        account_id,
+        bucket_id=None,
+        bucket_name=None,
+    ):
+        pass
+
+    @abstractmethod
+    def list_file_names(
+        self,
+        api_url,
+        account_auth_token,
+        bucket_id,
+        start_file_name=None,
+        max_file_count=None,
+        prefix=None,
+    ):
+        pass
+
+    @abstractmethod
+    def list_file_versions(
+        self,
+        api_url,
+        account_auth_token,
+        bucket_id,
+        start_file_name=None,
+        start_file_id=None,
+        max_file_count=None,
+        prefix=None,
+    ):
+        pass
+
+    @abstractmethod
+    def list_keys(
+        self,
+        api_url,
+        account_auth_token,
+        account_id,
+        max_key_count=None,
+        start_application_key_id=None
+    ):
+        pass
+
     @abstractmethod
     def list_parts(self, api_url, account_auth_token, file_id, start_part_number, max_part_count):
         pass
 
     @abstractmethod
     def list_unfinished_large_files(
-        self, api_url, account_auth_token, bucket_id, start_file_id=None, max_file_count=None
+        self,
+        api_url,
+        account_auth_token,
+        bucket_id,
+        start_file_id=None,
+        max_file_count=None,
+        prefix=None,
     ):
         pass
 
@@ -118,33 +247,25 @@ class AbstractRawApi(object):
     ):
         pass
 
+    @abstractmethod
+    def upload_file(
+        self, upload_url, upload_auth_token, file_name, content_length, content_type, content_sha1,
+        file_infos, data_stream
+    ):
+        pass
+
     @abstractmethod
     def upload_part(
         self, upload_url, upload_auth_token, part_number, content_length, sha1_sum, input_stream
     ):
         pass
 
-    def get_download_url_by_id(self, download_url, account_auth_token, file_id):
+    def get_download_url_by_id(self, download_url, file_id):
         return '%s/b2api/%s/b2_download_file_by_id?fileId=%s' % (download_url, API_VERSION, file_id)
 
-    def get_download_url_by_name(self, download_url, account_auth_token, bucket_name, file_name):
+    def get_download_url_by_name(self, download_url, bucket_name, file_name):
         return download_url + '/file/' + bucket_name + '/' + b2_url_encode(file_name)
 
-    @abstractmethod
-    def copy_file(
-        self,
-        api_url,
-        account_auth_token,
-        source_file_id,
-        new_file_name,
-        bytes_range=None,
-        metadata_directive=None,
-        content_type=None,
-        file_info=None,
-        destination_bucket_id=None,
-    ):
-        pass
-
 
 class B2RawApi(AbstractRawApi):
     """
@@ -161,10 +282,6 @@ class B2RawApi(AbstractRawApi):
     of the HTTP calls.  It can be mocked-out for testing higher layers.
     And this class can be tested by exercising each call just once,
     which is relatively quick.
-
-    All public methods of this class except authorize_account shall accept
-    api_url and account_info as first two positional arguments.  This is needed
-    for B2Session magic.
     """
 
     def __init__(self, b2_http):
@@ -256,11 +373,10 @@ class B2RawApi(AbstractRawApi):
             applicationKeyId=application_key_id,
         )
 
-    def download_file_from_url(self, _, account_auth_token_or_none, url, range_=None):
+    def download_file_from_url(self, account_auth_token_or_none, url, range_=None):
         """
         Issue a streaming request for download of a file, potentially authorized.
 
-        :param _: unused (caused by B2Session magic)
         :param account_auth_token_or_none: an optional account auth token to pass in
         :param url: the full URL to download from
         :param range: two-element tuple for http Range header
@@ -396,7 +512,13 @@ class B2RawApi(AbstractRawApi):
         )
 
     def list_unfinished_large_files(
-        self, api_url, account_auth_token, bucket_id, start_file_id=None, max_file_count=None
+        self,
+        api_url,
+        account_auth_token,
+        bucket_id,
+        start_file_id=None,
+        max_file_count=None,
+        prefix=None,
     ):
         return self._post_json(
             api_url,
@@ -404,7 +526,8 @@ class B2RawApi(AbstractRawApi):
             account_auth_token,
             bucketId=bucket_id,
             startFileId=start_file_id,
-            maxFileCount=max_file_count
+            maxFileCount=max_file_count,
+            namePrefix=prefix,
         )
 
     def start_large_file(
@@ -593,6 +716,30 @@ class B2RawApi(AbstractRawApi):
             **kwargs
         )
 
+    def copy_part(
+        self,
+        api_url,
+        account_auth_token,
+        source_file_id,
+        large_file_id,
+        part_number,
+        bytes_range=None,
+    ):
+        kwargs = {}
+        if bytes_range is not None:
+            range_dict = {}
+            _add_range_header(range_dict, bytes_range)
+            kwargs['range'] = range_dict['Range']
+        return self._post_json(
+            api_url,
+            'b2_copy_part',
+            account_auth_token,
+            sourceFileId=source_file_id,
+            largeFileId=large_file_id,
+            partNumber=part_number,
+            **kwargs
+        )
+
 
 def test_raw_api():
     """
@@ -695,7 +842,7 @@ def test_raw_api_helper(raw_api):
     print('b2_upload_file')
     file_name = 'test.txt'
     file_contents = six.b('hello world')
-    file_sha1 = hex_sha1_of_stream(six.BytesIO(file_contents), len(file_contents))
+    file_sha1 = hex_sha1_of_stream(io.BytesIO(file_contents), len(file_contents))
     file_dict = raw_api.upload_file(
         upload_url,
         upload_auth_token,
@@ -704,35 +851,35 @@ def test_raw_api_helper(raw_api):
         'text/plain',
         file_sha1,
         {'color': 'blue'},
-        six.BytesIO(file_contents),
+        io.BytesIO(file_contents),
     )
     file_id = file_dict['fileId']
 
     # b2_download_file_by_id with auth
     print('b2_download_file_by_id (auth)')
-    url = raw_api.get_download_url_by_id(download_url, None, file_id)
-    with raw_api.download_file_from_url(None, account_auth_token, url) as response:
+    url = raw_api.get_download_url_by_id(download_url, file_id)
+    with raw_api.download_file_from_url(account_auth_token, url) as response:
         data = next(response.iter_content(chunk_size=len(file_contents)))
         assert data == file_contents, data
 
     # b2_download_file_by_id no auth
     print('b2_download_file_by_id (no auth)')
-    url = raw_api.get_download_url_by_id(download_url, None, file_id)
-    with raw_api.download_file_from_url(None, None, url) as response:
+    url = raw_api.get_download_url_by_id(download_url, file_id)
+    with raw_api.download_file_from_url(None, url) as response:
         data = next(response.iter_content(chunk_size=len(file_contents)))
         assert data == file_contents, data
 
     # b2_download_file_by_name with auth
     print('b2_download_file_by_name (auth)')
-    url = raw_api.get_download_url_by_name(download_url, None, bucket_name, file_name)
-    with raw_api.download_file_from_url(None, account_auth_token, url) as response:
+    url = raw_api.get_download_url_by_name(download_url, bucket_name, file_name)
+    with raw_api.download_file_from_url(account_auth_token, url) as response:
         data = next(response.iter_content(chunk_size=len(file_contents)))
         assert data == file_contents, data
 
     # b2_download_file_by_name no auth
     print('b2_download_file_by_name (no auth)')
-    url = raw_api.get_download_url_by_name(download_url, None, bucket_name, file_name)
-    with raw_api.download_file_from_url(None, None, url) as response:
+    url = raw_api.get_download_url_by_name(download_url, bucket_name, file_name)
+    with raw_api.download_file_from_url(None, url) as response:
         data = next(response.iter_content(chunk_size=len(file_contents)))
         assert data == file_contents, data
 
@@ -745,8 +892,8 @@ def test_raw_api_helper(raw_api):
 
     # b2_download_file_by_name with download auth
     print('b2_download_file_by_name (download auth)')
-    url = raw_api.get_download_url_by_name(download_url, None, bucket_name, file_name)
-    with raw_api.download_file_from_url(None, download_auth_token, url) as response:
+    url = raw_api.get_download_url_by_name(download_url, bucket_name, file_name)
+    with raw_api.download_file_from_url(download_auth_token, url) as response:
         data = next(response.iter_content(chunk_size=len(file_contents)))
         assert data == file_contents, data
 
@@ -793,10 +940,10 @@ def test_raw_api_helper(raw_api):
     # b2_upload_part
     print('b2_upload_part')
     part_contents = six.b('hello part')
-    part_sha1 = hex_sha1_of_stream(six.BytesIO(part_contents), len(part_contents))
+    part_sha1 = hex_sha1_of_stream(io.BytesIO(part_contents), len(part_contents))
     raw_api.upload_part(
         upload_part_url, upload_path_auth, 1, len(part_contents), part_sha1,
-        six.BytesIO(part_contents)
+        io.BytesIO(part_contents)
     )
 
     # b2_list_parts
