@@ -9,8 +9,11 @@
 ######################################################################
 
 import collections
+import io
+import random
 import re
 import time
+import threading
 
 import six
 from six.moves import range
@@ -23,15 +26,23 @@ from .exception import (
     Conflict,
     DuplicateBucketName,
     FileNotPresent,
+    FileSha1Mismatch,
     InvalidAuthToken,
     InvalidMetadataDirective,
     MissingPart,
     NonExistentBucket,
+    PartSha1Mismatch,
     Unauthorized,
     UnsatisfiableRange,
 )
 from .raw_api import AbstractRawApi, HEX_DIGITS_AT_END, MetadataDirectiveMode
-from .utils import b2_url_decode, b2_url_encode
+from .utils import (
+    b2_url_decode,
+    b2_url_encode,
+    ConcurrentUsedAuthTokenGuard,
+    hex_sha1_of_bytes,
+)
+from .stream.hashing import StreamWithHash
 
 ALL_CAPABILITES = [
     'listKeys',
@@ -48,6 +59,19 @@ ALL_CAPABILITES = [
 ]
 
 
+def get_bytes_range(data_bytes, bytes_range):
+    """ Slice bytes array using bytes range """
+    if bytes_range is None:
+        return data_bytes
+    if bytes_range[0] > bytes_range[1]:
+        raise UnsatisfiableRange()
+    if bytes_range[0] < 0:
+        raise UnsatisfiableRange()
+    if bytes_range[1] > len(data_bytes):
+        raise UnsatisfiableRange()
+    return data_bytes[bytes_range[0]:bytes_range[1]]
+
+
 class KeySimulator(object):
     """
     Hold information about one application key, which can be either
@@ -283,7 +307,9 @@ class FakeResponse(object):
 
     def iter_content(self, chunk_size=1):
         start = 0
+        rnd = random.Random(self.url)
         while start <= len(self.data_bytes):
+            time.sleep(rnd.random() * 0.01)
             yield self.data_bytes[start:start + chunk_size]
             start += chunk_size
 
@@ -317,7 +343,8 @@ class BucketSimulator(object):
         bucket_type,
         bucket_info=None,
         cors_rules=None,
-        lifecycle_rules=None
+        lifecycle_rules=None,
+        options_set=None
     ):
         assert bucket_type in ['allPrivate', 'allPublic']
         self.account_id = account_id
@@ -327,6 +354,7 @@ class BucketSimulator(object):
         self.bucket_info = bucket_info or {}
         self.cors_rules = cors_rules or []
         self.lifecycle_rules = lifecycle_rules or []
+        self.options_set = options_set or set()
         self.revision = 1
         self.upload_url_counter = iter(range(200))
         # File IDs count down, so that the most recent will come first when they are sorted.
@@ -345,6 +373,7 @@ class BucketSimulator(object):
             bucketInfo=self.bucket_info,
             corsRules=self.cors_rules,
             lifecycleRules=self.lifecycle_rules,
+            options=self.options_set,
             revision=self.revision,
         )
 
@@ -398,7 +427,7 @@ class BucketSimulator(object):
         return dict(bucketId=self.bucket_id, uploadUrl=upload_url, authorizationToken=upload_url)
 
     def get_upload_part_url(self, file_id):
-        upload_url = 'https://upload.example.com/part/%s' % (file_id,)
+        upload_url = 'https://upload.example.com/part/%s/%d' % (file_id, random.randint(1, 10**9))
         return dict(bucketId=self.bucket_id, uploadUrl=upload_url, authorizationToken=upload_url)
 
     def hide_file(self, file_name):
@@ -437,14 +466,7 @@ class BucketSimulator(object):
         file_sim = self.file_id_to_file[file_id]
         new_file_id = self._next_file_id()
 
-        data_bytes = file_sim.data_bytes
-
-        if bytes_range is not None:
-            if bytes_range[0] >= len(file_sim.data_bytes
-                                    ) or bytes_range[1] <= 0:  # requested too much
-                raise UnsatisfiableRange()
-            else:
-                data_bytes = data_bytes[bytes_range[0]:bytes_range[1]]
+        data_bytes = get_bytes_range(file_sim.data_bytes, bytes_range)
 
         destination_bucket_id = destination_bucket_id or self.bucket_id
         copy_file_sim = self.FILE_SIMULATOR_CLASS(
@@ -469,6 +491,7 @@ class BucketSimulator(object):
         prev_file_name = None
         for key in sorted(six.iterkeys(self.file_name_and_id_to_file)):
             (file_name, file_id) = key
+            assert file_id
             if start_file_name <= file_name and file_name != prev_file_name:
                 if prefix is not None and not file_name.startswith(prefix):
                     break
@@ -498,8 +521,10 @@ class BucketSimulator(object):
         next_file_id = None
         for key in sorted(six.iterkeys(self.file_name_and_id_to_file)):
             (file_name, file_id) = key
-            if (start_file_name <
-                file_name) or (start_file_name == file_name and start_file_id <= file_id):
+            if (start_file_name < file_name) or (
+                start_file_name == file_name and
+                (start_file_id == '' or int(start_file_id) <= int(file_id))
+            ):
                 file_sim = self.file_name_and_id_to_file[key]
                 if prefix is not None and not file_name.startswith(prefix):
                     break
@@ -514,12 +539,13 @@ class BucketSimulator(object):
         file_sim = self.file_id_to_file[file_id]
         return file_sim.list_parts(start_part_number, max_part_count)
 
-    def list_unfinished_large_files(self, start_file_id=None, max_file_count=None):
+    def list_unfinished_large_files(self, start_file_id=None, max_file_count=None, prefix=None):
         start_file_id = start_file_id or self.FIRST_FILE_ID
         max_file_count = max_file_count or 100
         all_unfinished_ids = set(
             k for (k, v) in six.iteritems(self.file_id_to_file)
-            if v.action == 'start' and k <= start_file_id
+            if v.action == 'start' and k <= start_file_id and
+            (prefix is None or v.name.startswith(prefix))
         )
         ids_in_order = sorted(all_unfinished_ids, reverse=True)
         file_dict_list = [
@@ -576,12 +602,15 @@ class BucketSimulator(object):
         self, upload_id, upload_auth_token, file_name, content_length, content_type, content_sha1,
         file_infos, data_stream
     ):
-        data_bytes = data_stream.read()
+        data_bytes = self._simulate_chunked_post(data_stream, content_length)
         assert len(data_bytes) == content_length
         if content_sha1 == HEX_DIGITS_AT_END:
             content_sha1 = data_bytes[-40:].decode()
             data_bytes = data_bytes[0:-40]
             content_length -= 40
+        computed_sha1 = hex_sha1_of_bytes(data_bytes)
+        if content_sha1 != computed_sha1:
+            raise FileSha1Mismatch(file_name)
         file_id = self._next_file_id()
         file_sim = self.FILE_SIMULATOR_CLASS(
             self.account_id, self.bucket_id, file_id, 'upload', file_name, content_type,
@@ -593,12 +622,15 @@ class BucketSimulator(object):
 
     def upload_part(self, file_id, part_number, content_length, sha1_sum, input_stream):
         file_sim = self.file_id_to_file[file_id]
-        part_data = input_stream.read()
+        part_data = self._simulate_chunked_post(input_stream, content_length)
         assert len(part_data) == content_length
         if sha1_sum == HEX_DIGITS_AT_END:
             sha1_sum = part_data[-40:].decode()
             part_data = part_data[0:-40]
             content_length -= 40
+        computed_sha1 = hex_sha1_of_bytes(part_data)
+        if sha1_sum != computed_sha1:
+            raise PartSha1Mismatch(file_id)
         part = PartSimulator(file_sim.file_id, part_number, content_length, sha1_sum, part_data)
         file_sim.add_part(part_number, part)
         return dict(
@@ -608,6 +640,35 @@ class BucketSimulator(object):
             contentSha1=sha1_sum
         )  # yapf: disable
 
+    def _simulate_chunked_post(
+        self, stream, content_length, min_chunks=4, max_chunk_size=8096, simulate_retry=True
+    ):
+        chunk_size = max_chunk_size
+        chunks_num = self._chunks_number(content_length, chunk_size)
+        if chunks_num < min_chunks:
+            chunk_size = max(content_length // min_chunks, 1)
+        loop_count = 2 if simulate_retry else 1
+        stream_data = None
+        for _ in range(loop_count):
+            chunks = []
+            stream.seek(0)  # we always do this in `do_post` in `b2http` so we want it here *always*
+            while True:
+                data = stream.read(chunk_size)
+                chunks.append(data)
+                if not data:
+                    break
+            _stream_data = b''.join(chunks)
+            if stream_data is not None:
+                assert _stream_data == stream_data
+            stream_data = _stream_data
+        return stream_data
+
+    def _chunks_number(self, content_length, chunk_size):
+        chunks_number = content_length // chunk_size
+        if content_length % chunk_size > 0:
+            chunks_number = chunks_number + 1
+        return chunks_number
+
     def _next_file_id(self):
         return str(six.next(self.file_id_counter))
 
@@ -631,6 +692,7 @@ class RawSimulator(AbstractRawApi):
     MAX_DURATION_IN_SECONDS = 86400000
 
     UPLOAD_PART_MATCHER = re.compile('https://upload.example.com/part/([^/]*)')
+    UPLOAD_URL_MATCHER = re.compile(r'https://upload.example.com/([^/]*)/([^/]*)')
     DOWNLOAD_URL_MATCHER = re.compile(
         DOWNLOAD_URL + '(?:' + '|'.join(
             (
@@ -653,6 +715,10 @@ class RawSimulator(AbstractRawApi):
         # Set of auth tokens that have expired
         self.expired_auth_tokens = set()
 
+        # Map from auth token to a lock that upload procedure acquires
+        # when utilizing the token
+        self.currently_used_auth_tokens = collections.defaultdict(threading.Lock)
+
         # Counter for generating auth tokens.
         self.auth_token_counter = 0
 
@@ -821,7 +887,7 @@ class RawSimulator(AbstractRawApi):
         del self.bucket_id_to_bucket[bucket_id]
         return bucket.bucket_dict()
 
-    def download_file_from_url(self, _, account_auth_token_or_none, url, range_=None):
+    def download_file_from_url(self, account_auth_token_or_none, url, range_=None):
         # TODO: check auth token if bucket is not public
         matcher = self.DOWNLOAD_URL_MATCHER.match(url)
         assert matcher is not None, url
@@ -931,6 +997,33 @@ class RawSimulator(AbstractRawApi):
         dest_bucket.file_name_and_id_to_file[copy_file_sim.sort_key()] = copy_file_sim
         return copy_file_sim.as_upload_result()
 
+    def copy_part(
+        self,
+        api_url,
+        account_auth_token,
+        source_file_id,
+        large_file_id,
+        part_number,
+        bytes_range=None,
+    ):
+        src_bucket_id = self.file_id_to_bucket_id[source_file_id]
+        src_bucket = self._get_bucket_by_id(src_bucket_id)
+        dest_bucket_id = self.file_id_to_bucket_id[large_file_id]
+        dest_bucket = self._get_bucket_by_id(dest_bucket_id)
+
+        self._assert_account_auth(api_url, account_auth_token, dest_bucket.account_id, 'writeFiles')
+
+        file_sim = src_bucket.file_id_to_file[source_file_id]
+        data_bytes = get_bytes_range(file_sim.data_bytes, bytes_range)
+
+        data_stream = StreamWithHash(io.BytesIO(data_bytes), len(data_bytes))
+        content_length = len(data_stream)
+        sha1_sum = HEX_DIGITS_AT_END
+
+        return dest_bucket.upload_part(
+            large_file_id, part_number, content_length, sha1_sum, data_stream
+        )
+
     def list_buckets(
         self, api_url, account_auth_token, account_id, bucket_id=None, bucket_name=None
     ):
@@ -968,7 +1061,7 @@ class RawSimulator(AbstractRawApi):
     def list_file_names(
         self,
         api_url,
-        account_auth,
+        account_auth_token,
         bucket_id,
         start_file_name=None,
         max_file_count=None,
@@ -977,7 +1070,7 @@ class RawSimulator(AbstractRawApi):
         bucket = self._get_bucket_by_id(bucket_id)
         self._assert_account_auth(
             api_url,
-            account_auth,
+            account_auth_token,
             bucket.account_id,
             'listFiles',
             bucket_id=bucket_id,
@@ -988,7 +1081,7 @@ class RawSimulator(AbstractRawApi):
     def list_file_versions(
         self,
         api_url,
-        account_auth,
+        account_auth_token,
         bucket_id,
         start_file_name=None,
         start_file_id=None,
@@ -998,7 +1091,7 @@ class RawSimulator(AbstractRawApi):
         bucket = self._get_bucket_by_id(bucket_id)
         self._assert_account_auth(
             api_url,
-            account_auth,
+            account_auth_token,
             bucket.account_id,
             'listFiles',
             bucket_id=bucket_id,
@@ -1025,13 +1118,21 @@ class RawSimulator(AbstractRawApi):
         return bucket.list_parts(file_id, start_part_number, max_part_count)
 
     def list_unfinished_large_files(
-        self, api_url, account_auth, bucket_id, start_file_id=None, max_file_count=None
+        self,
+        api_url,
+        account_auth_token,
+        bucket_id,
+        start_file_id=None,
+        max_file_count=None,
+        prefix=None
     ):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth, bucket.account_id, 'listFiles')
+        self._assert_account_auth(
+            api_url, account_auth_token, bucket.account_id, 'listFiles', file_name=prefix
+        )
         start_file_id = start_file_id or ''
         max_file_count = max_file_count or 100
-        return bucket.list_unfinished_large_files(start_file_id, max_file_count)
+        return bucket.list_unfinished_large_files(start_file_id, max_file_count, prefix)
 
     def start_large_file(
         self, api_url, account_auth_token, bucket_id, file_name, content_type, file_info
@@ -1069,32 +1170,45 @@ class RawSimulator(AbstractRawApi):
         self, upload_url, upload_auth_token, file_name, content_length, content_type, content_sha1,
         file_infos, data_stream
     ):
-        assert upload_url == upload_auth_token
-        url_match = re.match(r'https://upload.example.com/([^/]*)/([^/]*)', upload_url)
-        if url_match is None:
-            raise BadUploadUrl(upload_url)
-        if len(self.upload_errors) != 0:
-            raise self.upload_errors.pop(0)
-        bucket_id, upload_id = url_match.groups()
-        bucket = self._get_bucket_by_id(bucket_id)
-        response = bucket.upload_file(
-            upload_id, upload_auth_token, file_name, content_length, content_type, content_sha1,
-            file_infos, data_stream
-        )  # yapf: disable
-        file_id = response['fileId']
-        self.file_id_to_bucket_id[file_id] = bucket_id
+        with ConcurrentUsedAuthTokenGuard(
+            self.currently_used_auth_tokens[upload_auth_token], upload_auth_token
+        ):
+            assert upload_url == upload_auth_token
+            url_match = self.UPLOAD_URL_MATCHER.match(upload_url)
+            if url_match is None:
+                raise BadUploadUrl(upload_url)
+            if self.upload_errors:
+                raise self.upload_errors.pop(0)
+            bucket_id, upload_id = url_match.groups()
+            bucket = self._get_bucket_by_id(bucket_id)
+            response = bucket.upload_file(
+                upload_id,
+                upload_auth_token,
+                file_name,
+                content_length,
+                content_type,
+                content_sha1,
+                file_infos,
+                data_stream,
+            )
+            file_id = response['fileId']
+            self.file_id_to_bucket_id[file_id] = bucket_id
         return response
 
     def upload_part(
         self, upload_url, upload_auth_token, part_number, content_length, sha1_sum, input_stream
     ):
-        url_match = self.UPLOAD_PART_MATCHER.match(upload_url)
-        if url_match is None:
-            raise BadUploadUrl(upload_url)
-        file_id = url_match.group(1)
-        bucket_id = self.file_id_to_bucket_id[file_id]
-        bucket = self._get_bucket_by_id(bucket_id)
-        return bucket.upload_part(file_id, part_number, content_length, sha1_sum, input_stream)
+        with ConcurrentUsedAuthTokenGuard(
+            self.currently_used_auth_tokens[upload_auth_token], upload_auth_token
+        ):
+            url_match = self.UPLOAD_PART_MATCHER.match(upload_url)
+            if url_match is None:
+                raise BadUploadUrl(upload_url)
+            file_id = url_match.group(1)
+            bucket_id = self.file_id_to_bucket_id[file_id]
+            bucket = self._get_bucket_by_id(bucket_id)
+            part = bucket.upload_part(file_id, part_number, content_length, sha1_sum, input_stream)
+        return part
 
     def _assert_account_auth(
         self, api_url, account_auth_token, account_id, capability, bucket_id=None, file_name=None
