@@ -77,12 +77,35 @@ class StubProgressListener(AbstractProgressListener):
         self.history.append('%d:' % (total_byte_count,))
 
     def bytes_completed(self, byte_count):
-        assert byte_count >= self.last_byte_count
         self.last_byte_count = byte_count
         self.history.append(str(byte_count))
 
-    def is_valid(self):
-        return self.total == self.last_byte_count
+    def is_valid(self, **kwargs):
+        valid, _ = self.is_valid_reason(**kwargs)
+        return valid
+
+    def is_valid_reason(
+        self, check_closed=True, check_progress=True, check_monotonic_progress=False
+    ):
+        progress_end = -1
+        if self.history[progress_end] == 'closed':
+            progress_end = -2
+
+        # self.total != self.last_byte_count may be a consequence of non-monotonic
+        # progress, so we want to check this first
+        if check_monotonic_progress:
+            prev = 0
+            for val in map(int, self.history[1:progress_end]):
+                if val < prev:
+                    return False, 'non-monotonic progress'
+                prev = val
+        if self.total != self.last_byte_count:
+            return False, 'total different than last_byte_count'
+        if check_closed and self.history[-1] != 'closed':
+            return False, 'no "closed" at the end of history'
+        if check_progress and len(self.history[1:progress_end]) < 2:
+            return False, 'progress in history has less than 2 entries'
+        return True, ''
 
     def close(self):
         self.history.append('closed')
@@ -156,15 +179,15 @@ class TestListParts(TestCaseWithBucket):
         content_sha1 = hex_sha1_of_bytes(content)
         large_file_upload_state = mock.MagicMock()
         large_file_upload_state.has_error.return_value = False
-        self.bucket._upload_part(
-            file1.file_id, 1, (0, 11), UploadSourceBytes(content), large_file_upload_state
-        )
-        self.bucket._upload_part(
-            file1.file_id, 2, (0, 11), UploadSourceBytes(content), large_file_upload_state
-        )
-        self.bucket._upload_part(
-            file1.file_id, 3, (0, 11), UploadSourceBytes(content), large_file_upload_state
-        )
+        self.api.services.upload_manager.upload_part(
+            self.bucket_id, file1.file_id, UploadSourceBytes(content), 1, large_file_upload_state
+        ).result()
+        self.api.services.upload_manager.upload_part(
+            self.bucket_id, file1.file_id, UploadSourceBytes(content), 2, large_file_upload_state
+        ).result()
+        self.api.services.upload_manager.upload_part(
+            self.bucket_id, file1.file_id, UploadSourceBytes(content), 3, large_file_upload_state
+        ).result()
         expected_parts = [
             Part('9999', 1, 11, content_sha1),
             Part('9999', 2, 11, content_sha1),
@@ -181,9 +204,10 @@ class TestUploadPart(TestCaseWithBucket):
         large_file_upload_state = LargeFileUploadState(file_progress_listener)
         large_file_upload_state.set_error('test error')
         try:
-            self.bucket._upload_part(
-                file1.file_id, 1, (0, 11), UploadSourceBytes(content), large_file_upload_state
-            )
+            self.bucket.api.services.upload_manager.upload_part(
+                self.bucket.id_, file1.file_id, UploadSourceBytes(content), 1,
+                large_file_upload_state
+            ).result()
             self.fail('should have thrown')
         except AlreadyFailed:
             pass
@@ -409,7 +433,7 @@ class TestUpload(TestCaseWithBucket):
         data = six.b('hello world')
         progress_listener = StubProgressListener()
         self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
-        self.assertEqual("11: 11 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_local_file(self):
         with TempDir() as d:
@@ -461,7 +485,7 @@ class TestUpload(TestCaseWithBucket):
         progress_listener = StubProgressListener()
         self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -472,7 +496,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_no_parts(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -482,7 +506,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertNotEqual(large_file_id, file_info.id_)  # it's not a match if there are no parts
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_all_parts_there(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -495,7 +519,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_part_does_not_match(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -506,7 +530,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_wrong_part_size(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -517,7 +541,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_file_info(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -530,7 +554,7 @@ class TestUpload(TestCaseWithBucket):
         )
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_file_info_does_not_match(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -543,7 +567,7 @@ class TestUpload(TestCaseWithBucket):
         )
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def _start_large_file(self, file_name, file_info=None):
         if file_info is None:
@@ -600,7 +624,12 @@ class DownloadTests(object):
     def _verify(self, expected_result, check_progress_listener=True):
         assert self.download_dest.get_bytes_written() == six.b(expected_result)
         if check_progress_listener:
-            assert self.progress_listener.is_valid()
+            valid, reason = self.progress_listener.is_valid_reason(
+                check_closed=False,
+                check_progress=False,
+                check_monotonic_progress=True,
+            )
+            assert valid, reason
 
     def test_download_by_id_no_progress(self):
         self.bucket.download_file_by_id(self.file_info.id_, self.download_dest)
@@ -729,13 +758,15 @@ class TestDownloadDefault(DownloadTests, EmptyFileDownloadScenarioMixin, TestCas
 class TestDownloadSimple(DownloadTests, EmptyFileDownloadScenarioMixin, TestCaseWithBucket):
     def setUp(self):
         super(TestDownloadSimple, self).setUp()
-        self.bucket.api.transferer.strategies = [SimpleDownloader(force_chunk_size=20,)]
+        self.bucket.api.services.download_manager.strategies = [
+            SimpleDownloader(force_chunk_size=20,)
+        ]
 
 
 class TestDownloadParallel(DownloadTests, TestCaseWithBucket):
     def setUp(self):
         super(TestDownloadParallel, self).setUp()
-        self.bucket.api.transferer.strategies = [
+        self.bucket.api.services.download_manager.strategies = [
             ParallelDownloader(
                 force_chunk_size=2,
                 max_streams=999,
@@ -776,13 +807,15 @@ class TestCaseWithTruncatedDownloadBucket(TestCaseWithBucket):
 class TestTruncatedDownloadSimple(DownloadTests, TestCaseWithTruncatedDownloadBucket):
     def setUp(self):
         super(TestTruncatedDownloadSimple, self).setUp()
-        self.bucket.api.transferer.strategies = [SimpleDownloader(force_chunk_size=20,)]
+        self.bucket.api.services.download_manager.strategies = [
+            SimpleDownloader(force_chunk_size=20,)
+        ]
 
 
 class TestTruncatedDownloadParallel(DownloadTests, TestCaseWithTruncatedDownloadBucket):
     def setUp(self):
         super(TestTruncatedDownloadParallel, self).setUp()
-        self.bucket.api.transferer.strategies = [
+        self.bucket.api.services.download_manager.strategies = [
             ParallelDownloader(
                 force_chunk_size=3,
                 max_streams=2,
