@@ -21,12 +21,14 @@ from .test_base import TestBase
 from .deps_exception import (
     AlreadyFailed,
     B2Error,
+    B2RequestTimeoutDuringUpload,
     InvalidAuthToken,
     InvalidMetadataDirective,
     InvalidRange,
     InvalidUploadSource,
     MaxRetriesExceeded,
     UnsatisfiableRange,
+    FileSha1Mismatch,
 )
 from .deps import B2Api
 from .deps import LargeFileUploadState
@@ -77,12 +79,35 @@ class StubProgressListener(AbstractProgressListener):
         self.history.append('%d:' % (total_byte_count,))
 
     def bytes_completed(self, byte_count):
-        assert byte_count >= self.last_byte_count
         self.last_byte_count = byte_count
         self.history.append(str(byte_count))
 
-    def is_valid(self):
-        return self.total == self.last_byte_count
+    def is_valid(self, **kwargs):
+        valid, _ = self.is_valid_reason(**kwargs)
+        return valid
+
+    def is_valid_reason(
+        self, check_closed=True, check_progress=True, check_monotonic_progress=False
+    ):
+        progress_end = -1
+        if self.history[progress_end] == 'closed':
+            progress_end = -2
+
+        # self.total != self.last_byte_count may be a consequence of non-monotonic
+        # progress, so we want to check this first
+        if check_monotonic_progress:
+            prev = 0
+            for val in map(int, self.history[1:progress_end]):
+                if val < prev:
+                    return False, 'non-monotonic progress'
+                prev = val
+        if self.total != self.last_byte_count:
+            return False, 'total different than last_byte_count'
+        if check_closed and self.history[-1] != 'closed':
+            return False, 'no "closed" at the end of history'
+        if check_progress and len(self.history[1:progress_end]) < 2:
+            return False, 'progress in history has less than 2 entries'
+        return True, ''
 
     def close(self):
         self.history.append('closed')
@@ -156,15 +181,15 @@ class TestListParts(TestCaseWithBucket):
         content_sha1 = hex_sha1_of_bytes(content)
         large_file_upload_state = mock.MagicMock()
         large_file_upload_state.has_error.return_value = False
-        self.bucket._upload_part(
-            file1.file_id, 1, (0, 11), UploadSourceBytes(content), large_file_upload_state
-        )
-        self.bucket._upload_part(
-            file1.file_id, 2, (0, 11), UploadSourceBytes(content), large_file_upload_state
-        )
-        self.bucket._upload_part(
-            file1.file_id, 3, (0, 11), UploadSourceBytes(content), large_file_upload_state
-        )
+        self.api.services.upload_manager.upload_part(
+            self.bucket_id, file1.file_id, UploadSourceBytes(content), 1, large_file_upload_state
+        ).result()
+        self.api.services.upload_manager.upload_part(
+            self.bucket_id, file1.file_id, UploadSourceBytes(content), 2, large_file_upload_state
+        ).result()
+        self.api.services.upload_manager.upload_part(
+            self.bucket_id, file1.file_id, UploadSourceBytes(content), 3, large_file_upload_state
+        ).result()
         expected_parts = [
             Part('9999', 1, 11, content_sha1),
             Part('9999', 2, 11, content_sha1),
@@ -181,9 +206,10 @@ class TestUploadPart(TestCaseWithBucket):
         large_file_upload_state = LargeFileUploadState(file_progress_listener)
         large_file_upload_state.set_error('test error')
         try:
-            self.bucket._upload_part(
-                file1.file_id, 1, (0, 11), UploadSourceBytes(content), large_file_upload_state
-            )
+            self.api.services.upload_manager.upload_part(
+                self.bucket_id, file1.file_id, UploadSourceBytes(content), 1,
+                large_file_upload_state
+            ).result()
             self.fail('should have thrown')
         except AlreadyFailed:
             pass
@@ -205,6 +231,18 @@ class TestListUnfinished(TestCaseWithBucket):
             [file1, file2, file3], list(self.bucket.list_unfinished_large_files(batch_size=1))
         )
 
+    def test_prefix(self):
+        self.bucket.start_large_file('fileA', 'text/plain', {})
+        file2 = self.bucket.start_large_file('fileAB', 'text/plain', {})
+        file3 = self.bucket.start_large_file('fileABC', 'text/plain', {})
+        self.assertEqual(
+            [file2, file3],
+            list(self.bucket.list_unfinished_large_files(
+                batch_size=1,
+                prefix='fileAB',
+            ),),
+        )
+
     def _make_file(self, file_id, file_name):
         return self.bucket.start_large_file(file_name, 'text/plain', {})
 
@@ -290,6 +328,81 @@ class TestLs(TestCaseWithBucket):
         self.assertBucketContents(expected, '', show_versions=True)
 
 
+class TestListVersions(TestCaseWithBucket):
+    def test_single_version(self):
+        data = six.b('hello world')
+        a_id = self.bucket.upload_bytes(data, 'a').id_
+        b_id = self.bucket.upload_bytes(data, 'b').id_
+        c_id = self.bucket.upload_bytes(data, 'c').id_
+
+        expected = [(a_id, 'a', 11, 'upload')]
+        actual = [
+            (info.id_, info.file_name, info.size, info.action)
+            for info in self.bucket.list_file_versions('a')
+        ]
+        self.assertEqual(expected, actual)
+
+        expected = [(b_id, 'b', 11, 'upload')]
+        actual = [
+            (info.id_, info.file_name, info.size, info.action)
+            for info in self.bucket.list_file_versions('b')
+        ]
+        self.assertEqual(expected, actual)
+
+        expected = [(c_id, 'c', 11, 'upload')]
+        actual = [
+            (info.id_, info.file_name, info.size, info.action)
+            for info in self.bucket.list_file_versions('c')
+        ]
+        self.assertEqual(expected, actual)
+
+    def test_multiple_version(self):
+        a_id1 = self.bucket.upload_bytes(six.b('first version'), 'a').id_
+        a_id2 = self.bucket.upload_bytes(six.b('second version'), 'a').id_
+        a_id3 = self.bucket.upload_bytes(six.b('last version'), 'a').id_
+
+        expected = [
+            (a_id3, 'a', 12, 'upload'), (a_id2, 'a', 14, 'upload'), (a_id1, 'a', 13, 'upload')
+        ]
+        actual = [
+            (info.id_, info.file_name, info.size, info.action)
+            for info in self.bucket.list_file_versions('a')
+        ]
+        self.assertEqual(expected, actual)
+
+    def test_ignores_subdirectory(self):
+        data = six.b('hello world')
+        file_id = self.bucket.upload_bytes(data, 'a/b').id_
+        self.bucket.upload_bytes(data, 'a/b/c')
+
+        expected = [(file_id, 'a/b', 11, 'upload')]
+        actual = [
+            (info.id_, info.file_name, info.size, info.action)
+            for info in self.bucket.list_file_versions('a/b')
+        ]
+        self.assertEqual(expected, actual)
+
+    def test_all_versions_in_response(self):
+        data = six.b('hello world')
+        file_id = self.bucket.upload_bytes(data, 'a/b').id_
+        self.bucket.upload_bytes(data, 'a/b/c')
+
+        expected = [(file_id, 'a/b', 11, 'upload')]
+        actual = [
+            (info.id_, info.file_name, info.size, info.action)
+            for info in self.bucket.list_file_versions('a/b', fetch_count=1)
+        ]
+        self.assertEqual(expected, actual)
+
+    def test_bad_fetch_count(self):
+        try:
+            # Convert to a list to cause the generator to execute.
+            list(self.bucket.list_file_versions('a', fetch_count=0))
+            self.fail('should have raised ValueError')
+        except ValueError as e:
+            self.assertEqual('unsupported fetch_count value', str(e))
+
+
 class TestCopyFile(TestCaseWithBucket):
     def test_copy_without_optional_params(self):
         file_id = self._make_file()
@@ -406,7 +519,7 @@ class TestUpload(TestCaseWithBucket):
         data = six.b('hello world')
         progress_listener = StubProgressListener()
         self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
-        self.assertEqual("11: 11 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_local_file(self):
         with TempDir() as d:
@@ -434,11 +547,28 @@ class TestUpload(TestCaseWithBucket):
             with self.assertRaises(InvalidUploadSource):
                 self.bucket.upload_local_file(path, 'file1')
 
+    def test_upload_local_wrong_sha(self):
+        with TempDir() as d:
+            path = os.path.join(d, 'file123')
+            data = six.b('hello world')
+            write_file(path, data)
+            with self.assertRaises(FileSha1Mismatch):
+                self.bucket.upload_local_file(
+                    path,
+                    'file123',
+                    sha1_sum='abcabcabc',
+                )
+
     def test_upload_one_retryable_error(self):
         self.simulator.set_upload_errors([CanRetry(True)])
         data = six.b('hello world')
         self.bucket.upload_bytes(data, 'file1')
 
+    def test_upload_timeout(self):
+        self.simulator.set_upload_errors([B2RequestTimeoutDuringUpload()])
+        data = six.b('hello world')
+        self.bucket.upload_bytes(data, 'file1')
+
     def test_upload_file_one_fatal_error(self):
         self.simulator.set_upload_errors([CanRetry(False)])
         data = six.b('hello world')
@@ -456,7 +586,15 @@ class TestUpload(TestCaseWithBucket):
         progress_listener = StubProgressListener()
         self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
+
+    def test_upload_local_large_file(self):
+        with TempDir() as d:
+            path = os.path.join(d, 'file1')
+            data = self._make_data(self.simulator.MIN_PART_SIZE * 3)
+            write_file(path, data)
+            self.bucket.upload_local_file(path, 'file1')
+            self._check_file_contents('file1', data)
 
     def test_upload_large_resume(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -467,7 +605,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_no_parts(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -477,7 +615,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertNotEqual(large_file_id, file_info.id_)  # it's not a match if there are no parts
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_all_parts_there(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -490,7 +628,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_part_does_not_match(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -501,7 +639,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_wrong_part_size(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -512,7 +650,7 @@ class TestUpload(TestCaseWithBucket):
         file_info = self.bucket.upload_bytes(data, 'file1', progress_listener=progress_listener)
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_file_info(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -525,7 +663,7 @@ class TestUpload(TestCaseWithBucket):
         )
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
 
     def test_upload_large_resume_file_info_does_not_match(self):
         part_size = self.simulator.MIN_PART_SIZE
@@ -538,7 +676,19 @@ class TestUpload(TestCaseWithBucket):
         )
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
-        self.assertEqual("600: 200 400 600 closed", progress_listener.get_history())
+        self.assertTrue(progress_listener.is_valid())
+
+    def test_upload_large_file_with_restricted_api_key(self):
+        self.simulator.key_id_to_key[self.account_id].name_prefix_or_none = 'path/to'
+        part_size = self.simulator.MIN_PART_SIZE
+        data = self._make_data(part_size * 3)
+        progress_listener = StubProgressListener()
+        file_info = self.bucket.upload_bytes(
+            data, 'path/to/file1', progress_listener=progress_listener
+        )
+        self.assertEqual(len(data), file_info.size)
+        self._check_file_contents('path/to/file1', data)
+        self.assertTrue(progress_listener.is_valid())
 
     def _start_large_file(self, file_name, file_info=None):
         if file_info is None:
@@ -583,11 +733,11 @@ class TestUpload(TestCaseWithBucket):
 # Downloads
 
 
-class DownloadTests(object):
-    DATA = 'abcdefghijklmnopqrs'
+class DownloadTestsBase(object):
+    DATA = NotImplemented
 
     def setUp(self):
-        super(DownloadTests, self).setUp()
+        super(DownloadTestsBase, self).setUp()
         self.file_info = self.bucket.upload_bytes(six.b(self.DATA), 'file1')
         self.download_dest = DownloadDestBytes()
         self.progress_listener = StubProgressListener()
@@ -595,7 +745,16 @@ class DownloadTests(object):
     def _verify(self, expected_result, check_progress_listener=True):
         assert self.download_dest.get_bytes_written() == six.b(expected_result)
         if check_progress_listener:
-            assert self.progress_listener.is_valid()
+            valid, reason = self.progress_listener.is_valid_reason(
+                check_closed=False,
+                check_progress=False,
+                check_monotonic_progress=True,
+            )
+            assert valid, reason
+
+
+class DownloadTests(DownloadTestsBase):
+    DATA = 'abcdefghijklmnopqrs'
 
     def test_download_by_id_no_progress(self):
         self.bucket.download_file_by_id(self.file_info.id_, self.download_dest)
@@ -724,21 +883,46 @@ class TestDownloadDefault(DownloadTests, EmptyFileDownloadScenarioMixin, TestCas
 class TestDownloadSimple(DownloadTests, EmptyFileDownloadScenarioMixin, TestCaseWithBucket):
     def setUp(self):
         super(TestDownloadSimple, self).setUp()
-        self.bucket.api.transferer.strategies = [SimpleDownloader(force_chunk_size=20,)]
+        self.bucket.api.services.download_manager.strategies = [
+            SimpleDownloader(force_chunk_size=20,)
+        ]
 
 
 class TestDownloadParallel(DownloadTests, TestCaseWithBucket):
     def setUp(self):
         super(TestDownloadParallel, self).setUp()
-        self.bucket.api.transferer.strategies = [
+        self.bucket.api.services.download_manager.strategies = [
             ParallelDownloader(
                 force_chunk_size=2,
                 max_streams=999,
                 min_part_size=2,
-            )
+            ),
         ]
 
 
+class TestDownloadParallelALotOfStreams(DownloadTestsBase, TestCaseWithBucket):
+    DATA = ''.join(['01234567890abcdef'] * 32)
+
+    def setUp(self):
+        super(TestDownloadParallelALotOfStreams, self).setUp()
+        self.bucket.api.services.download_manager.strategies = [
+            # this should produce 32 streams with 16 single byte writes
+            # so we increase probability of non-sequential writes as much as possible
+            # with great help from random sleeps in FakeResponse
+            ParallelDownloader(
+                force_chunk_size=1,
+                max_streams=32,
+                min_part_size=16,
+            ),
+        ]
+
+    def test_download_by_id_progress_monotonic(self):
+        self.bucket.download_file_by_id(
+            self.file_info.id_, self.download_dest, self.progress_listener
+        )
+        self._verify(self.DATA)
+
+
 # Truncated downloads
 
 
@@ -771,13 +955,15 @@ class TestCaseWithTruncatedDownloadBucket(TestCaseWithBucket):
 class TestTruncatedDownloadSimple(DownloadTests, TestCaseWithTruncatedDownloadBucket):
     def setUp(self):
         super(TestTruncatedDownloadSimple, self).setUp()
-        self.bucket.api.transferer.strategies = [SimpleDownloader(force_chunk_size=20,)]
+        self.bucket.api.services.download_manager.strategies = [
+            SimpleDownloader(force_chunk_size=20,)
+        ]
 
 
 class TestTruncatedDownloadParallel(DownloadTests, TestCaseWithTruncatedDownloadBucket):
     def setUp(self):
         super(TestTruncatedDownloadParallel, self).setUp()
-        self.bucket.api.transferer.strategies = [
+        self.bucket.api.services.download_manager.strategies = [
             ParallelDownloader(
                 force_chunk_size=3,
                 max_streams=2,
